use std::marker::PhantomData;

use anyhow::Error;
use async_trait::async_trait;
use diesel::prelude::*;
use diesel::associations::HasTable;
use diesel::query_dsl::methods::LoadQuery;

use crate::config::db::DbPool;
use super::i_service::IBaseService;

pub struct BaseService<T, U>
where
    T: Identifiable<Id = i32> + Queryable<U::SqlType, diesel::pg::Pg> + Send + Sync + 'static,
    U: Table + HasTable<Table = U> + LoadQuery<'static, PgConnection, T> + Send + Sync + 'static,
{
    pool: DbPool,
    _marker: PhantomData<(T, U)>,
}

impl<T, U> BaseService<T, U>
where
    T: Identifiable<Id = i32> + Queryable<U::SqlType, diesel::pg::Pg> + Send + Sync + 'static,
    U: Table 
        + HasTable<Table = U> 
        + LoadQuery<'static, PgConnection, T>
        + Send 
        + Sync 
        + 'static,
{
    pub fn new(pool: DbPool) -> Self {
        Self {
            pool,
            _marker: PhantomData,
        }
    }
}

#[async_trait]
impl<T, U> IBaseService<T> for BaseService<T, U>
where
    T: Identifiable<Id = i32> + Queryable<U::SqlType, diesel::pg::Pg> + Send + Sync + 'static,
    U: Table 
        + HasTable<Table = U> 
        + LoadQuery<'static, PgConnection, T>
        + Send 
        + Sync 
        + 'static,
{
    async fn get_all(&self) -> Result<Vec<T>, Error> {
        let pool = self.pool.clone();

        let result = tokio::task::spawn_blocking(move || {
            let mut conn = pool.get().expect("Failed to get DB connection");
            U::table().load::<T>(&mut conn)
        })
        .await??;

        Ok(result)
    }
}